/**
 * Created on 2008-03-17 21:25:07
 */
package org.put.semintec.core;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import org.semanticweb.kaon2.api.Axiom;
import org.semanticweb.kaon2.api.DefaultOntologyResolver;
import org.semanticweb.kaon2.api.Entity;
import org.semanticweb.kaon2.api.KAON2Connection;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.Ontology;
import org.semanticweb.kaon2.api.OntologyChangeEvent;
import org.semanticweb.kaon2.api.Request;
import org.semanticweb.kaon2.api.formatting.OntologyFileFormat;
import org.semanticweb.kaon2.api.owl.axioms.ClassMember;
import org.semanticweb.kaon2.api.owl.axioms.DataPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.Declaration;
import org.semanticweb.kaon2.api.owl.axioms.DifferentIndividuals;
import org.semanticweb.kaon2.api.owl.axioms.NegativeDataPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.NegativeObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.ObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.SameIndividual;
import org.semanticweb.kaon2.api.owl.elements.Individual;
import org.semanticweb.kaon2.api.reasoner.Reasoner;

/**
 * Represents a knowledge base on which mining algorithm is run. 
 * @author Agnieszka Lawrynowicz
 */
public class KnowledgeBase {

/*
 * (non-javadoc)
 */
private DefaultOntologyResolver resolver;
 
/**
 * Getter of the property <tt>resolver</tt>
 *
 * @return Returns the resolver.
 * 
 */
public DefaultOntologyResolver getResolver()
{
	return resolver;
}

/**
 * Setter of the property <tt>resolver</tt>
 *
 * @param resolver The resolver to set.
 *
 */
public void setResolver(DefaultOntologyResolver resolver ){
	this.resolver = resolver;
}

/*
 * (non-javadoc)
 */
private KAON2Connection connection;
 
/**
 * Getter of the property <tt>connection</tt>
 *
 * @return Returns the connection.
 * 
 */
public KAON2Connection getConnection()
{
	return connection;
}

/**
 * Setter of the property <tt>connection</tt>
 *
 * @param connection The connection to set.
 *
 */
public void setConnection(KAON2Connection connection ){
	this.connection = connection;
}

	/**
 * Reasoner associated with given knowledge base.
 */
private Reasoner reasoner;
 
/**
 * Getter of the property <tt>r</tt>
 *
 * @return Returns the reasoner r.
 * 
 */
public Reasoner getReasoner()
{
	return reasoner;
}

/**
 * Setter of the property <tt>r</tt>
 *
 * @param r The reasoner to set.
 *
 */
public void setReasoner(Reasoner r ){
	this.reasoner = r;
}

	/**
 * Ontology that is represented by the knowledge base.
 */
private Ontology ontology;
 

/**
 * Getter of the property <tt>o</tt>
 *
 * @return Returns the ontology.
 * 
 */
public Ontology getOntology()
{
	return ontology;
}

/**
 * Setter of the property <tt>o</tt>
 *
 * @param ontology The ontology to set.
 *
 */
public void setOntology(Ontology ontology ){
	this.ontology = ontology;
}

	/**
	 * Makes a copy of a given knowledge base.
	 * @param includeABox True if a TBox and ABox should be copied, false if only a TBox should be copied.
	 * @return A copy of a given knowledge base.
	 * @throws InterruptedException 
	 * @throws IOException 
	 * @throws KAON2Exception 
	*/
	public KnowledgeBase copy(boolean includeABox) throws KAON2Exception, IOException, InterruptedException{
		
		KnowledgeBase kbCopy = new KnowledgeBase();
		
        //DefaultOntologyResolver resolver=new DefaultOntologyResolver();
        File tmpFile = File.createTempFile("semintec", "");
        
        resolver.registerReplacement("http://"+tmpFile.getName(),tmpFile.toURI().toString());
        //connection.setOntologyResolver(resolver);

		Ontology oCopy = connection.createOntology("http://"+tmpFile.getName(),new HashMap<String,Object>());
		Reasoner rCopy = oCopy.createReasoner();
	
		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
        
		Request<Axiom> axiomRequest=ontology.createAxiomRequest();

		Set<Axiom> axioms = axiomRequest.get();
		for (Axiom axiom : axioms)
		{
	//		System.out.println(axiom);

			if (axiom instanceof ClassMember || 
				axiom instanceof ObjectPropertyMember ||
				axiom instanceof NegativeObjectPropertyMember ||
				axiom instanceof DataPropertyMember ||
				axiom instanceof NegativeDataPropertyMember ||
				axiom instanceof DifferentIndividuals ||
				axiom instanceof SameIndividual ||
				(axiom instanceof Declaration && ((Declaration) axiom).getEntity() instanceof Individual)) 
				 
			{
				if (includeABox)
				{
					changes.add(new OntologyChangeEvent(axiom,OntologyChangeEvent.ChangeType.ADD));
				} 
		    }
			else
			{
				changes.add(new OntologyChangeEvent(axiom,OntologyChangeEvent.ChangeType.ADD));
			}
		}
		
		oCopy.applyChanges(changes);
		
		kbCopy.setOntology(oCopy);
		kbCopy.setReasoner(rCopy);
		
		return kbCopy;
	}
	
}
