/**
 * Created on 2008-03-18 12:01:55
 */
package org.put.semintec.core;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import jdsl.core.api.Position;
import jdsl.core.api.Tree;

import org.put.semintec.mining.association.Bias;
import org.put.semintec.mining.association.Pattern;
import org.put.semintec.mining.association.PatternMining;
import org.put.semintec.mining.association.AssociationRule;
import org.put.semintec.mining.association.AssociationRuleGeneration;
import org.put.semintec.mining.datastructures.Node;
import org.put.semintec.mining.datastructures.Trie;
import org.put.semintec.utils.IOUtils;
import org.put.semintec.utils.Statistics;
import org.put.semintec.utils.KBUtils;
import org.put.semintec.utils.MiningUtils;

import org.semanticweb.kaon2.api.Axiom;
import org.semanticweb.kaon2.api.Cursor;
import org.semanticweb.kaon2.api.DefaultOntologyResolver;
import org.semanticweb.kaon2.api.Entity;
import org.semanticweb.kaon2.api.KAON2Connection;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.Namespaces;
import org.semanticweb.kaon2.api.OntologyChangeEvent;
import org.semanticweb.kaon2.api.formatting.OntologyFileFormat;
import org.semanticweb.kaon2.api.logic.Conjunction;
import org.semanticweb.kaon2.api.logic.Disjunction;
import org.semanticweb.kaon2.api.logic.Formula;
import org.semanticweb.kaon2.api.logic.Literal;
import org.semanticweb.kaon2.api.logic.Predicate;
import org.semanticweb.kaon2.api.logic.PredicateSymbol;
import org.semanticweb.kaon2.api.logic.QueryDefinition;
import org.semanticweb.kaon2.api.logic.Rule;
import org.semanticweb.kaon2.api.logic.Term;
import org.semanticweb.kaon2.api.logic.Variable;
import org.semanticweb.kaon2.api.owl.axioms.ClassMember;
import org.semanticweb.kaon2.api.owl.axioms.DataPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.Declaration;
import org.semanticweb.kaon2.api.owl.axioms.DifferentIndividuals;
import org.semanticweb.kaon2.api.owl.axioms.NegativeDataPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.NegativeObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.ObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.SameIndividual;
import org.semanticweb.kaon2.api.owl.elements.Description;
import org.semanticweb.kaon2.api.owl.elements.Individual;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;
import org.semanticweb.kaon2.api.owl.elements.ObjectOr;
import org.semanticweb.kaon2.api.owl.elements.ObjectProperty;
import org.semanticweb.kaon2.api.reasoner.Query;
import org.semanticweb.kaon2.api.Request;


/**
 * Represents an example application that executes SEMINTEC algorithms.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class SemintecApp {

	/**
	 * @param args setupFileName
	 * @throws KAON2Exception 
	 * @throws InterruptedException 
	 * @throws IOException 
	 */
	public static void main(String[] args) throws KAON2Exception, InterruptedException, IOException {
        IOUtils ioUtils = new IOUtils();
        HashMap params = ioUtils.readXMLSetupFile(args[0]);
        
		KnowledgeBase kb = new KnowledgeBase();

		KAON2Connection connection=KAON2Manager.newConnection();
		DefaultOntologyResolver resolver=new DefaultOntologyResolver();

		resolver.registerReplacement((String)params.get("ontologyURI"), (String)params.get("physicalURI"));
        connection.setOntologyResolver(resolver);

        kb.setConnection(connection);
        kb.setResolver(resolver);
        kb.setOntology(connection.openOntology((String)params.get("ontologyURI"),new HashMap<String,Object>()));
        kb.setReasoner(kb.getOntology().createReasoner());

//Example of how to model disjunctive rules and Datalog facts-----
//		OWLClass Client = KAON2Manager.factory().owlClass("http://www.owl-ontologies.com/ex.owl#Client");
//		PredicateSymbol man = KAON2Manager.factory().predicateSymbol("http://www.owl-ontologies.com/ex.owl#p_man", 1);
//		PredicateSymbol woman = KAON2Manager.factory().predicateSymbol("http://www.owl-ontologies.com/ex.owl#p_woman", 1);
//		Individual anna = KAON2Manager.factory().individual("http://www.owl-ontologies.com/ex.owl#Anna");
		
//		Variable x1 = KAON2Manager.factory().variable("x1");

//		Literal client_x1=KAON2Manager.factory().literal(true,Client,new Term[] { x1 });
//		Literal woman_x1=KAON2Manager.factory().literal(true,woman,new Term[] { x1 });
//		Literal man_x1=KAON2Manager.factory().literal(true,man,new Term[] { x1 });

//		Literal woman_anna=KAON2Manager.factory().literal(true,woman,new Term[] { anna });
		
//		Set ss = new HashSet();
//		ss.add(woman_x1);
//		ss.add(man_x1);
		
//		Disjunction d = KAON2Manager.factory().disjunction(ss);
		
//		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
//		Rule r = KAON2Manager.factory().rule(
//				d,                          // this is the rule head, i.e. the consequent of the rule
//				new Literal[] {client_x1}   // this is the rule body, i.e. the condition of the rule
//		);

//		changes.add(new OntologyChangeEvent(r ,OntologyChangeEvent.ChangeType.ADD));
//		changes.add(new OntologyChangeEvent(woman_anna ,OntologyChangeEvent.ChangeType.ADD));

//		kb.getOntology().applyChanges(changes);
//		kb.getOntology().saveOntology(OntologyFileFormat.FUNCTIONAL_SYNTAX,new File("c:\\temp\\testSemFact.xml"),"ISO-8859-1");
// end of disjunctive rule definition example        
        
		KnowledgeBase kbCopy = kb.copy(false);

		OWLClass refClass=KAON2Manager.factory().owlClass((String)params.get("refclass"));
	    
		Bias bias = new Bias();
		
		ArrayList predlist = (ArrayList)params.get("predlist");
		
		Request<Entity> entityRequest = kb.getOntology().createEntityRequest();
		Set<Entity> entities = entityRequest.get();
		OWLClass thing = KAON2Manager.factory().thing();
		OWLClass nothing = KAON2Manager.factory().nothing();
		
		for (Entity entity : entities) {
			if (entity instanceof OWLClass || entity instanceof ObjectProperty || entity instanceof PredicateSymbol)
			{
					if (predlist.isEmpty() ||
							predlist.contains(entity.toString()) ){
						Query q = kb.getReasoner().createQuery((Predicate)entity);
						q.open();
						if (!q.afterLast()) {
							bias.predicates.add(entity);
						}
						q.close();
						q.dispose();
					}
			} 
		}

   		bias.predicates.add(refClass);
   		
		Statistics statistics = new Statistics();
		statistics.setMiningTask(Statistics.PATTERN);

		PatternMining pm = new PatternMining();
        pm.setKb(kb);
        pm.setKbCopy(kbCopy);
        pm.setMinSupport(Double.valueOf((String)params.get("minsup")).doubleValue());
        pm.setMaxLength(Integer.valueOf((String)params.get("maxlength")).intValue());
        pm.setRefClass(refClass);
        pm.setBias(bias);
        pm.setStatistics(statistics);
        if (((String)params.get("answercaching")).compareTo("yes") == 0) {
        	pm.setAnswerCaching(true);
        } else {
        	pm.setAnswerCaching(false);
        }
        pm.getStatistics().startToCollect();
        pm.mine();
        pm.getStatistics().stopToCollect();
        
        ioUtils.saveToGraphML(pm.getTrie(), (String)params.get("patternsFileName")+"_dl.graphml", new String[1]);
        ioUtils.saveStatistics(params, statistics, (String)params.get("statisticsFileName")+"_dl");
        ioUtils.savePatterns(params, pm.getTrie(), (String)params.get("patternsFileName")+"_dl");

        AssociationRuleGeneration arg = new AssociationRuleGeneration();
        arg.setTrie(pm.getTrie());
        arg.setKbCopy(kbCopy);
        arg.generate();
        
        ioUtils.saveRules(params, arg.getRules(), (String)params.get("rulesFileName"));

        kb.getReasoner().dispose();
        kbCopy.getReasoner().dispose();
        connection.close();
	}
	
}
