/**
 * Created on 2008-03-17 21:25:55  
 */
package org.put.semintec.mining.association;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import org.put.semintec.core.Atom;
import org.put.semintec.core.KnowledgeBase;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.OntologyChangeEvent;
import org.semanticweb.kaon2.api.formatting.OntologyFileFormat;
import org.semanticweb.kaon2.api.logic.Conjunction;
import org.semanticweb.kaon2.api.logic.FactFormula;
import org.semanticweb.kaon2.api.logic.Formula;
import org.semanticweb.kaon2.api.logic.Literal;
import org.semanticweb.kaon2.api.logic.Predicate;
import org.semanticweb.kaon2.api.logic.PredicateSymbol;
import org.semanticweb.kaon2.api.logic.QueryDefinition;
import org.semanticweb.kaon2.api.logic.Term;
import org.semanticweb.kaon2.api.logic.Variable;
import org.semanticweb.kaon2.api.owl.axioms.ClassMember;
import org.semanticweb.kaon2.api.owl.axioms.ObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.elements.Individual;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;
import org.semanticweb.kaon2.api.owl.elements.ObjectProperty;
import org.semanticweb.kaon2.api.reasoner.Query;


/**
 * Represents a pattern.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class Pattern {
	/**
	 * Conjunctive query definition of the query that is represented by the pattern.
	 */
	private QueryDefinition queryDefinition;
	 
	/**
	 * Getter of the property <tt>queryDefinition</tt>
	 *
	 * @return Returns the queryDefinition.
	 * 
	 */
	public QueryDefinition getQueryDefinition()
	{
		return queryDefinition;
	}


	/**
	 * Setter of the property <tt>queryDefinition</tt>
	 *
	 * @param queryDefinition The queryDefinition to set.
	 *
	 */
	public void setQueryDefinition(QueryDefinition queryDefinition ){
		this.queryDefinition = queryDefinition;
	}

	/**
	 * Number of literals in the pattern.
	 */
	//private int length;
	 
	/**
	 * Getter of the property <tt>length</tt>
	 *
	 * @return Returns the length.
	 * 
	 */
	//public int getLength()
	//{
	//	return length;
	//}

	/**
	 * Setter of the property <tt>length</tt>
	 *
	 * @param length The length to set.
	 *
	 */
	//public void setLength(int length){
	//	this.length = length;
	//}
	
/**
	 * @param QueryDefinition associated with the pattern 
	*/
	public Pattern(QueryDefinition qd){
		setQueryDefinition(qd);
		undistinguishedVariables = new ArrayList();
	}


/*
 * Stores the value of the key variable in the query associated with the pattern
 *  after skolemization.
 */
private Individual skolemizedKeyVar;
 
/**
 * Getter of the property <tt>skolemizedKeyVar</tt>
 *
 * @return Returns the skolemizedKeyVar.
 * 
 */
private Individual getSkolemizedKeyVar()
{
	return skolemizedKeyVar;
}

/**
 * Setter of the property <tt>skolemizedKeyVar</tt>
 *
 * @param skolemizedKeyVar The skolemizedKeyVar to set.
 *
 */
private void setSkolemizedKeyVar(Individual skolemizedKeyVar ){
	this.skolemizedKeyVar = skolemizedKeyVar;
}

/**
 * Undistinguished variables in a query associated with the given pattern.
 */
private ArrayList undistinguishedVariables;
 
/**
 * Getter of the property <tt>undistinguishedVariables</tt>
 *
 * @return Returns the undistinguishedVariables.
 * 
 */
public ArrayList getUndistinguishedVariables()
{
	return undistinguishedVariables;
}

/**
 * Setter of the property <tt>undistinguishedVariables</tt>
 *
 * @param undistinguishedVariables The undistinguishedVariables to set.
 *
 */
public void setUndistinguishedVariables(ArrayList undistinguishedVariables ){
	this.undistinguishedVariables = undistinguishedVariables;
}

/**
 * Number of answers of the query represented by the given pattern.
 */
private int frequency;
 
/**
 * Getter of the property <tt>frequency</tt>
 *
 * @return Returns the frequency.
 * 
 */
public int getFrequency()
{
	return frequency;
}

/**
 * Setter of the property <tt>frequency</tt>
 *
 * @param frequency The frequency to set.
 *
 */
public void setFrequency(int frequency ){
	this.frequency = frequency;
}

/**
 * Support of the pattern calculated w.r.t. its frequency and frequency of the reference class
 */
private double support;
 
/**
 * Getter of the property <tt>support</tt>
 *
 * @return Returns the support.
 * 
 */
public double getSupport()
{
	return support;
}

/**
 * Setter of the property <tt>support</tt>
 *
 * @param support The support to set.
 *
 */
public void setSupport(double support ){
	this.support = support;
}

/**
 * Refines pattern by adding an atom to it
 * @param a Atom by which current pattern is extended
 * @return New refined pattern 
* @throws KAON2Exception 
*/
public Pattern refine(Atom a) throws KAON2Exception {
	ArrayList<Literal> literals = new ArrayList<Literal>();
	for (Formula f : ((Conjunction) queryDefinition.getFormula()).getConjuncts()) {
		if (f instanceof Literal){
			literals.add((Literal)f);
		}
	}

	literals.add(a.getAtom());
	
	Formula conj = KAON2Manager.factory().conjunction(literals);
	
	QueryDefinition newQueryDefinition = KAON2Manager.factory().queryDefinition(conj,
    		getQueryDefinition().getDistinguishedVariables()); 
    
	Pattern p = new Pattern(newQueryDefinition);
	//tmp
	//newQuery.dispose();
	
	return p;
} 

/**
 * @param p A pattern that is tested whether it is semantically equivalent to the given pattern.
 * @param kb A knowledge base w.r.t. pattern equivalence is tested. 
 * @return True if the given pattern is equivalent to pattern <tt>p</tt> w.r.t. a knowledge base <tt>kb</tt>.   
 * @throws InterruptedException 
 * @throws KAON2Exception 
*/
public boolean equivalentTo(Pattern p, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
	if (this.subsumedBy(p, kb) && p.subsumedBy(this, kb)){
		return true;
	}
	else {
		return false;
	}
}

/**
	 * Checks if the pattern is semantically free w.r.t. given knowledge base,
	 *  that is if it is not possible to delete any of its atoms
	 *  without affecting the semantics.
	 *  The exception to this procedure is the atom with the reference concept 
	 *  which is not included in s-freeness tests. 
	 *  @param kb A knowledge base
	 *  @return True if the pattern is semantically free  
 * @throws InterruptedException 
 * @throws KAON2Exception 
 * @throws KAON2Exception 
 * @throws InterruptedException 
	*/
	public boolean isSFree(KnowledgeBase kb) throws KAON2Exception, InterruptedException {
		boolean result = true;
		int i = 0;
		
		//removes atom with reference concept from s-freeness tests
		ArrayList lits = new ArrayList();
		ArrayList literals = (ArrayList)((Conjunction) getQueryDefinition().getFormula()).getConjuncts();
		lits.addAll(literals);
		literals = null;
		lits.remove(0);
		
		Variable keyVar = getQueryDefinition().getDistinguishedVariables()[0];

		ArrayList vars = new ArrayList();
		vars.add(keyVar);
		//Variable[] vars = new Variable[1];
		//vars[0]= keyVar;
		
		Formula fLits = KAON2Manager.factory().conjunction(lits);
		
	    //builds a pattern without reference/key atom
		QueryDefinition childQueryDefinition =  KAON2Manager.factory().queryDefinition(fLits, vars); 
	    Pattern childPattern = new Pattern(childQueryDefinition);
		
	    //tmp
	    //childQuery.dispose();
	    //removes each literal and tests if shorter pattern is not more 
		//general than longer one
		while (i<lits.size() && result)
		{
			ArrayList newLits = new ArrayList();
			newLits.addAll(lits);
			newLits.remove(i);
			
			//TODO - make new, separate method with this test
			//check if new query is range-restricted
			//it should be always range-restricted, because of the method of generation
			//remove this test
			boolean rangeRestricted = false;
			int j = 0;
			while (!rangeRestricted && j<newLits.size()) {
				if (((Literal) newLits.get(j)).containsVariable(keyVar)) {
					rangeRestricted = true;
				}
				j++;
			}
			
			if (rangeRestricted){
		    	//builds a pattern without some literal
				Formula fNewLits = KAON2Manager.factory().conjunction(newLits);

				QueryDefinition parentQueryDefinition = KAON2Manager.factory().queryDefinition(fNewLits, vars);// kb.getReasoner().createQuery(newLits, vars); 
				Pattern parentPattern = new Pattern(parentQueryDefinition);
				//tmp
				//parentQuery.dispose();
		
				if (parentPattern.subsumedBy(childPattern, kb)) {
					result = false;
				}
				//parentPattern.getQuery().dispose();
				parentPattern = null;
		}
			
			i++;
		}
		
		//childPattern.getQuery().dispose();
		//childPattern = null;
		
		return result;
	}


	/**
	 * 	Skolemizes a query associated with the given pattern -- replaces all variables with new constants
     *  thus creating new ABox
	 * @param kb A knowledge base to store result of the skolemization
	 * @return A list of ontology change events (of type <OntologyChangeEvent>) 
	 * to rollback the skolemization from the knowledge base 
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	 * 
	*/
	public List skolemize(KnowledgeBase kb) throws KAON2Exception, InterruptedException
	{
		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
	    List<OntologyChangeEvent> rollback=new ArrayList<OntologyChangeEvent>();

		ArrayList literals = (ArrayList)((Conjunction) queryDefinition.getFormula()).getConjuncts();

		for (Iterator it=literals.iterator(); it.hasNext();)
		{
			Literal l = (Literal)it.next();
			ArrayList inds = new ArrayList();
    	
			for (int j=0; j<l.getArity(); j++)
			{
				Variable v = (Variable) l.getArgument(j);
				Individual ind=KAON2Manager.factory().individual(kb.getOntology().getOntologyURI()+"#test"+v.getVariableName());
				
				//TODO - sprawdzanie czy juz takiego nie ma, jesli jest to inna nazwa
				inds.add(ind);
				if (v.equals(queryDefinition.getDistinguishedVariables()[0]))
				{
					setSkolemizedKeyVar(ind);
				}
			}
    	
			Predicate p = l.getPredicate();
			if (p instanceof OWLClass)
			{
				OWLClass c = (OWLClass) p;
				ClassMember cm = KAON2Manager.factory().classMember(c, (Individual)inds.get(0));
		    
				if (!kb.getOntology().containsAxiom(cm,false))
				{
					changes.add(new OntologyChangeEvent(KAON2Manager.factory().classMember(c,(Individual) inds.get(0)),OntologyChangeEvent.ChangeType.ADD));
					kb.getOntology().applyChanges(changes);
					changes.clear();
					rollback.add(new OntologyChangeEvent(KAON2Manager.factory().classMember(c,(Individual) inds.get(0)),OntologyChangeEvent.ChangeType.REMOVE));
				}
			}
			else if (p instanceof ObjectProperty)
			{
				ObjectProperty op = (ObjectProperty) p;
			
				ObjectPropertyMember opm = KAON2Manager.factory().objectPropertyMember(op,(Individual)inds.get(0),(Individual)inds.get(1));
				if (!kb.getOntology().containsAxiom(opm,false))
				{
					changes.add(new OntologyChangeEvent(KAON2Manager.factory().objectPropertyMember(op,(Individual)inds.get(0),(Individual)inds.get(1)),OntologyChangeEvent.ChangeType.ADD));
					rollback.add(new OntologyChangeEvent(KAON2Manager.factory().objectPropertyMember(op,(Individual)inds.get(0),(Individual)inds.get(1)),OntologyChangeEvent.ChangeType.REMOVE));
					kb.getOntology().applyChanges(changes);
					changes.clear();

				}
			}
			else if (p instanceof PredicateSymbol)
			{
				PredicateSymbol ps = (PredicateSymbol) p;
    		
				Literal lit = KAON2Manager.factory().literal(true, ps, inds);
				if (!kb.getOntology().containsAxiom(lit,false))
				{
					changes.add(new OntologyChangeEvent(KAON2Manager.factory().literal(true, ps, inds),OntologyChangeEvent.ChangeType.ADD));
					rollback.add(new OntologyChangeEvent(KAON2Manager.factory().literal(true, ps, inds),OntologyChangeEvent.ChangeType.REMOVE));
					kb.getOntology().applyChanges(changes);
					changes.clear();
				}
			}		
		}

//		kb.getOntology().saveOntology(OntologyFileFormat.OWL_RDF,new File("c:\\temp\\testSem.xml"),"ISO-8859-1");
		
		return rollback;
	}
	
	/**
	 * 
	 * @param kb A knowledge base
	 * @return True if a pattern is satisfiable w.r.t. a given knowledge base.   
	 * @throws KAON2Exception 
	 * @throws InterruptedException 
	*/
	public boolean isSatisfiable(KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		
	    //skolemization of query variables: replacing all variables with new constants
	    //Q(a,b) = new ABox
	    List<OntologyChangeEvent> rollback = skolemize(kb);
	    
	    //check whether kb is satisfiable in conjunction with new ABox
	    boolean result = kb.getReasoner().isSatisfiable();
	    
	    kb.getOntology().applyChanges(rollback);
        rollback.clear();
	    
	    return result;
	}

/**
 * @param p A pattern that is tested whether it subsumes the given pattern.
 * @param kb A knowledge base w.r.t. pattern subsumption is tested. 
 * @return True if the given pattern is subsumed by pattern <tt>p</tt> w.r.t. a knowledge base <tt>kb</tt>.   
 * @throws KAON2Exception 
 * @throws InterruptedException 
*/
	public boolean subsumedBy(Pattern p, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
	    List<OntologyChangeEvent> rollback = skolemize(kb);
	    boolean result = false;
	    Query newQuery = kb.getReasoner().createQuery(p.getQueryDefinition());
	    		
	    newQuery.open();
	    if (newQuery.getNumberOfTuples()>0)
	    {
	    	Term term[] = newQuery.tupleBuffer();
	    	Individual i = (Individual) term[0];
	    	if (i.equals(getSkolemizedKeyVar()))
	    	{
	    		result = true;
	    	}
	    	else
	    	{
	    		result = false;
	    	}
	    }
	    else
	    {
	    	result = false;
	    }
	    newQuery.close();
	    newQuery.dispose();

	    kb.getOntology().applyChanges(rollback);
        rollback.clear();
        setSkolemizedKeyVar(null);
        
		return result;
	}

	/**
	 * Evaluates pattern, that is calculates its support
	 * @param refClassFrequency Frequency of the reference class wrt which the support of the pattern is calculated
	 * @param kb Knowledge base wrt which pattern should be evaluated
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	 * 
	*/
	public void evaluate(int refClassFrequency, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		Query q = kb.getReasoner().createQuery(getQueryDefinition());
		q.open();
		setFrequency(q.getNumberOfTuples());
		q.close();
		q.dispose();
		
		setSupport((double) ((double)frequency)/(double)(refClassFrequency));
	}
	
	
	/**
	 * @param minSupp Minimum support threshold
	 * @return True if the support of the pattern is above minimum support threshold false otherwise 
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	 * 
	*/
	public boolean isFrequent(double minSupp) throws KAON2Exception, InterruptedException{
		if (support > minSupp){
			return true;
		} else {
			return false;
		}
	}

	/**
	 * @return Undistinguished variables of the query associated with the pattern.
	*/
	public ArrayList computeUndistinguishedVariables() {
		ArrayList vars = new ArrayList();
		
		for (Formula f : ((Conjunction) getQueryDefinition().getFormula()).getConjuncts()) {
			if (f instanceof Literal){
				for (int j=0; j<((Literal) f).getArity(); j++) {
					Variable v = (Variable) ((Literal) f).getArgumentVariable(j);
					if (!v.equals(getQueryDefinition().getDistinguishedVariables()[0]) && !vars.contains(v)) {
						vars.add(v);
					}
				}
			}
		}
		return vars;
	}

	/**
	 * @param a Atom  with the variables from which the resulting list should be generated
	 * @return Undistinguished variables of the query associated with the pattern that 
	 * appear in both: the query and in atom a
	*/
	public ArrayList computeUndistinguishedVariablesSharedWith(Atom a) {
		ArrayList vars = new ArrayList();
		
		for (Formula f : ((Conjunction) getQueryDefinition().getFormula()).getConjuncts()) {
			if (f instanceof Literal){
				for (int j=0; j<((Literal) f).getArity(); j++) {
					Variable v = (Variable) ((Literal) f).getArgumentVariable(j);
					if (!v.equals(getQueryDefinition().getDistinguishedVariables()[0]) &&
						a.getAtom().containsVariable(v)	&&
						!vars.contains(v)) {
						vars.add(v);
					}
				}
			}
		}
		return vars;
	}

	
	/**
	 * Rewrites query Q1 associated with the given pattern into new query Q2 
	 * where all variables of Q1 (distinguished and undistinguished ones) are distinguished in Q2 
	 * @return New pattern, based on query Q2, where all variables from the initial pattern 
	 * (query Q1) are distinguished ones
	 *  
	 * @throws KAON2Exception 
	*/
	public Pattern rewriteVariablesToDistinguishedOnes() throws KAON2Exception{
		ArrayList<Variable> vars = new ArrayList<Variable>();
		
		//add reference variable
		vars.add(getQueryDefinition().getDistinguishedVariables()[0]);
		//add remaining variables
		vars.addAll(computeUndistinguishedVariables());
		
		QueryDefinition qDefinitionDistinguished = KAON2Manager.factory().queryDefinition(getQueryDefinition().getFormula(), vars);
	    
		Pattern p = new Pattern(qDefinitionDistinguished);
		return p;
	}

	
	/**
	 * Determines, according to principles of SMARTCALL transformation, bindings of what variables will be affected by adding new atom 
	 * (except the reference variable)
	 * @param lastAtom Last atom of the pattern   
	 * @return List of variables whose bindings will be affected by adding new atom
	 *  
	 * @throws KAON2Exception 
	*/
	public ArrayList determineVariablesConnectedToLastAtom(Atom lastAtom) throws KAON2Exception{
		ArrayList<Variable> varsConnectedToLastAtom = new ArrayList<Variable>();

		//add reference variable
		for (int i=0; i<lastAtom.getAtom().getArity();i++){
			if (!lastAtom.getAtom().getArgumentVariable(i).equals(getQueryDefinition().getDistinguishedVariables()[0])){
				varsConnectedToLastAtom.add(lastAtom.getAtom().getArgumentVariable(i));
			}
		}
		
		ArrayList literals = new ArrayList();
		for (Formula f : ((Conjunction) getQueryDefinition().getFormula()).getConjuncts()) {
			if (f instanceof Literal){
				literals.add(f);
			}
		}
		literals.remove(lastAtom.getAtom());

		int size = 0;

		while (varsConnectedToLastAtom.size()!=size){
			size = varsConnectedToLastAtom.size();

			for (Iterator ii = literals.iterator(); ii.hasNext(); ){
				Literal l =((Literal)ii.next());
				for (int i=0; i<l.getArity();i++){
					Variable v = l.getArgumentVariable(i);
					if (!varsConnectedToLastAtom.contains(v) && 
						!v.equals(getQueryDefinition().getDistinguishedVariables()[0])) {
						boolean containsAny = false;
						int j=0; 
						while (j<l.getArity() && !containsAny){
							if (varsConnectedToLastAtom.contains(l.getArgumentVariable(j))){
								containsAny = true;
							}
							j++;
						}
						if (containsAny){
							varsConnectedToLastAtom.add(v);
						}
					}
				}
			}
		}
		
		return varsConnectedToLastAtom;
	}
	


}
