/**
 * Created on 2008-03-18 11:58:34
 */
package org.put.semintec.mining.datastructures;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import jdsl.core.api.Position;
import org.put.semintec.core.Atom;
import org.put.semintec.core.KnowledgeBase;
import org.put.semintec.mining.association.Pattern;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.OntologyChangeEvent;
import org.semanticweb.kaon2.api.logic.FactFormula;
import org.semanticweb.kaon2.api.logic.Formula;
import org.semanticweb.kaon2.api.logic.Literal;
import org.semanticweb.kaon2.api.logic.PredicateSymbol;
import org.semanticweb.kaon2.api.logic.QueryDefinition;
import org.semanticweb.kaon2.api.logic.Rule;
import org.semanticweb.kaon2.api.logic.Term;
import org.semanticweb.kaon2.api.logic.Variable;
import org.semanticweb.kaon2.api.owl.axioms.ClassMember;
import org.semanticweb.kaon2.api.owl.axioms.ObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.elements.Individual;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;
import org.semanticweb.kaon2.api.owl.elements.ObjectProperty;
import org.semanticweb.kaon2.api.reasoner.Query;

/**
 * Represents a node in a trie data structure.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class Node {
	
	public Node(){
		lastMaterializationDepth = new HashMap();
	}

	/**
 * Stores hash map with variables as keys and as value last depth were the variable bindings were materialized
 */
private HashMap lastMaterializationDepth;
 
/**
 * Getter of the property <tt>lastMaterializationDepth</tt>
 *
 * @return Returns the lastMaterializationDepth.
 * 
 */
public HashMap getLastMaterializationDepth()
{
	return lastMaterializationDepth;
}

/**
 * Setter of the property <tt>lastMaterializationDepth</tt>
 *
 * @param lastMaterializationDepth The lastMaterializationDepth to set.
 *
 */
public void setLastMaterializationDepth(HashMap lastMaterializationDepth ){
	this.lastMaterializationDepth = lastMaterializationDepth;
}

	/**
 * Pattern rewritten in such way, that instead normal literals (except last one), it contains special ones for each variable. Used in answer caching technique.
 */
private Pattern rewrittenPattern;
 
/**
 * Getter of the property <tt>rewrittenPattern</tt>
 *
 * @return Returns the rewrittenPattern.
 * 
 */
public Pattern getRewrittenPattern()
{
	return rewrittenPattern;
}

/**
 * Setter of the property <tt>rewrittenPattern</tt>
 *
 * @param rewrittenPattern The rewrittenPattern to set.
 *
 */
public void setRewrittenPattern(Pattern rewrittenPattern ){
	this.rewrittenPattern = rewrittenPattern;
}

	public String how;
	
	/*
 * Depth of the node in a trie
 */
private int depth;
 
/**
 * Getter of the property <tt>depth</tt>
 *
 * @return Returns the depth.
 * 
 */
public int getDepth()
{
	return depth;
}

/**
 * Setter of the property <tt>depth</tt>
 *
 * @param depth The depth to set.
 *
 */
public void setDepth(int depth ){
	this.depth = depth;
}



	/**
 * Pattern that is represented by trie node
 */
private Pattern pattern;
 
/**
 * Getter of the property <tt>pattern</tt>
 *
 * @return Returns the pattern.
 * 
 */
public Pattern getPattern()
{
	return pattern;
}

/**
 * Setter of the property <tt>pattern</tt>
 *
 * @param pattern The pattern to set.
 *
 */
public void setPattern(Pattern pattern ){
	this.pattern = pattern;
}

/**
 * Position of the node in ordered tree associated with given trie
 */
private Position pos;
 
/**
 * Getter of the property <tt>pos</tt>
 *
 * @return Returns the pos.
 * 
 */
public Position getPos()
{
	return pos;
}

/**
 * Setter of the property <tt>pos</tt>
 *
 * @param pos The pos to set.
 *
 */
public void setPos(Position pos ){
	this.pos = pos;
}

/*
 * (non-javadoc)
 */
private boolean materialized = false;
 
/**
 * Getter of the property <tt>marked</tt>
 *
 * @return Returns the marked.
 * 
 */
public boolean getMaterialized()
{
	return materialized;
}

/**
 * Setter of the property <tt>marked</tt>
 *
 * @param marked The marked to set.
 *
 */
public void setMaterialized(boolean materialized ){
	this.materialized = materialized;
}

/**
 * Stores variable index in an atom associated with given node
 * True values are where there are variables introduced to a query
 * by the atom  (new variables).
 */
private boolean[] newVarPos;
 
/**
 * Getter of the property <tt>newVars</tt>
 *
 * @return Returns the newVars.
 * 
 */
public boolean[] getNewVarPos()
{
	return newVarPos;
}

/**
 * Setter of the property <tt>newVars</tt>
 *
 * @param newVars The newVars to set.
 *
 */
public void setNewVarPos(boolean[] newVarPos ){
	this.newVarPos = newVarPos;
}

	/**
 * An atom represented by the node.
 */
private Atom atom;
 
/**
 * Getter of the property <tt>atom</tt>
 *
 * @return Returns the atom.
 * 
 */
public Atom getAtom()
{
	return atom;
}

/**
 * Setter of the property <tt>atom</tt>
 *
 * @param atom The atom to set.
 *
 */
public void setAtom(Atom atom ){
	this.atom = atom;
}

/**
 * Updates hash map that stores variables as keys and as value last depth were the variable
 *  bindings were materialized
 *  @param vars List of variables that were materialized at the current depth
 */

private	void updateLastMaterializationDepth(ArrayList vars){
	for (Iterator i=vars.iterator(); i.hasNext();){
		Variable v = (Variable) i.next();
		getLastMaterializationDepth().put(v, new Integer(getDepth()));
	}
}


/**
 * Rewrites query Q1 associated with the given node into new query Q2 
 * where literals of Q1 are rewritten to special predicates
 * in Q2 (one per each variable appearing in both: query Q1 and atom a). 
 * The special predicates are to store cached answers.
 * @param parent node of the given pattern
 * @param kb Knowledge base where special predicates are stored  
 * @return New pattern, based on query Q2 and with last atom from current node
 *  
 * @throws KAON2Exception 
*/
public Pattern rewriteLiteralsToSupOnes(Node parent, KnowledgeBase kb) throws KAON2Exception{
	//distinguished variables
	ArrayList<Variable> vars = new ArrayList<Variable>();
	ArrayList<Literal> literals = new ArrayList<Literal>();
	
	//add reference variable
	vars.add(getPattern().getQueryDefinition().getDistinguishedVariables()[0]);

	//undistinguished variables
	ArrayList uVars = new ArrayList();
	uVars.addAll(parent.getPattern().computeUndistinguishedVariablesSharedWith(getAtom()));

	//add special predicate for reference variable
	if (getAtom().getAtom().containsVariable((Variable)vars.get(0))){
		OWLClass refMaterializedClass =  KAON2Manager.factory().owlClass(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(parent.getDepth()));
//		PredicateSymbol refMaterializedClass =  KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(parent.getDepth()),1);
		
		Literal l = KAON2Manager.factory().literal(true, refMaterializedClass, new Variable[]{(Variable)vars.get(0)});
		literals.add(l);
	}
	
	//add special predicates for remaining variables (undistinguished ones)
	for (int j = 0; j < uVars.size(); j++) {
		Variable v = (Variable)uVars.get(j);
	    ObjectProperty materializedProperty = KAON2Manager.factory().objectProperty(kb.getOntology().getOntologyURI()+"#$sup_"+String.valueOf(parent.getLastMaterializationDepth().get(v))+"_"+v.toString());
		//PredicateSymbol materializedProperty = KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$sup_"+String.valueOf(parent.getLastMaterializationDepth().get(v))+"_"+v.toString(),2);

		Literal l = KAON2Manager.factory().literal(true, materializedProperty, new Variable[]{(Variable)vars.get(0), v});
		literals.add(l);
	}

	//add last atom
	literals.add(getAtom().getAtom());
	Formula f = KAON2Manager.factory().conjunction(literals);
	
	QueryDefinition qDefinitionRewritten = KAON2Manager.factory().queryDefinition(f, vars);
    
	Pattern p = new Pattern(qDefinitionRewritten);
	return p;
}

/**
 * Creates rewritten pattern to be materialized, which is based on the original pattern, 
 * and has distinguished variables specified as input, and it has
 * literals with supporting predicates for all old/input variables of the original pattern
 *  and last atom in the body
 *	@param kb Knowledge base from which URI is taken for the pattern 
 *  @param vars List of variables which bindings are to be materialized at the current depth
 *  (distinguished variables of the created pattern)
 */

private	Pattern createRewrittenPattern(ArrayList vars, KnowledgeBase kb){
	ArrayList uVars =  getPattern().computeUndistinguishedVariables();
	ArrayList lits = new ArrayList();		
	ArrayList newVars = new ArrayList();
	
	for (int i = 0; i<getNewVarPos().length;i++){
		if (getNewVarPos()[i]){
			newVars.add(getAtom().getAtom().getArgumentVariable(i));
		}
	}
	
	OWLClass refMaterializedClass =  KAON2Manager.factory().owlClass(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(getDepth()-1));
	//PredicateSymbol refMaterializedClass =  KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(getDepth()-1),1);
	
	
	Literal l = KAON2Manager.factory().literal(true, refMaterializedClass, new Variable[]{getPattern().getQueryDefinition().getDistinguishedVariable(0)});
	lits.add(l);
	
	for (Iterator i=uVars.iterator(); i.hasNext();){
		Variable v = (Variable) i.next();
		//if variable is not new (what means, it was materialized earlier)
		if (!newVars.contains(v)){
			ObjectProperty op = KAON2Manager.factory().objectProperty(kb.getOntology().getOntologyURI()+"#$sup_"+String.valueOf(getLastMaterializationDepth().get(v))+"_"+v);
			//PredicateSymbol op = KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$sup_"+String.valueOf(getLastMaterializationDepth().get(v))+"_"+v, 2);
			Literal lit = KAON2Manager.factory().literal(true, op, new Variable[]{getPattern().getQueryDefinition().getDistinguishedVariable(0), v});
			lits.add(lit);
		}
	}

	lits.add(getAtom().getAtom());
	
	Formula f = KAON2Manager.factory().conjunction(lits);
	QueryDefinition qd = KAON2Manager.factory().queryDefinition(f, vars);
	Pattern p = new Pattern(qd);
	return p;
}

/**
 * Materializes answers of rewritten query associated with the node - for each relevant variable 
 * creates special predicates to store answers.
 * @param kb Knowledge base to which answers are to be materialized
 * @return List of changes (of type<OntologyChangeEvent>) to knowledge base, made by adding materialized answers, to rollback  
 * @throws InterruptedException 
 * @throws KAON2Exception 
 * 
*/
public List materializeResults(KnowledgeBase kb) throws KAON2Exception, InterruptedException{
	//Pattern p = getRewrittenPattern().rewriteVariablesToDistinguishedOnes();

	ArrayList connectedVars  = getPattern().determineVariablesConnectedToLastAtom(getAtom());
	
	ArrayList vars = new ArrayList();
	vars.add(getPattern().getQueryDefinition().getDistinguishedVariable(0));
	vars.addAll(connectedVars);
	
	Pattern p = createRewrittenPattern(vars, kb);
	
	OWLClass refMaterializedClass =  KAON2Manager.factory().owlClass(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(getDepth()));
	//PredicateSymbol refMaterializedClass =  KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(getDepth()),1);

	
	List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
	List<OntologyChangeEvent> rollback=new ArrayList<OntologyChangeEvent>();

	Query q = kb.getReasoner().createQuery(p.getQueryDefinition());
	q.open();

	
	while (!q.afterLast()) {
		Term t[] = q.tupleBuffer();
		
		Individual refMaterializedIndividual = (Individual) t[0];

		ClassMember cm = KAON2Manager.factory().classMember(refMaterializedClass,refMaterializedIndividual);
	///	Literal ll =  KAON2Manager.factory().literal(true, refMaterializedClass,refMaterializedIndividual);
		//Formula cm = KAON2Manager.factory().literal(true, refMaterializedClass,refMaterializedIndividual);
	//	Rule cm = KAON2Manager.factory().rule(
		//		ll,                          // this is the rule head, i.e. the consequent of the rule
	      //      new Literal[] { }   // this is the rule body, i.e. the condition of the rule
	       // );

		if (!kb.getOntology().containsAxiom(cm,false)) {
//			changes.add(new OntologyChangeEvent(KAON2Manager.factory().classMember(refMaterializedClass,refMaterializedIndividual) ,OntologyChangeEvent.ChangeType.ADD));
			changes.add(new OntologyChangeEvent(cm, OntologyChangeEvent.ChangeType.ADD));
			
			kb.getOntology().applyChanges(changes);
			changes.clear();
				
			rollback.add(new OntologyChangeEvent(cm,OntologyChangeEvent.ChangeType.REMOVE));
		}
		
		for (int k=1; k<t.length; k++) {
			Individual i = (Individual) t[k];
			
		   //PredicateSymbol ps =   KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$sup_"+String.valueOf(getDepth())+"_"+p.getQueryDefinition().getDistinguishedVariables()[k],2);
			ObjectProperty op = KAON2Manager.factory().objectProperty(kb.getOntology().getOntologyURI()+"#$sup_"+String.valueOf(getDepth())+"_"+p.getQueryDefinition().getDistinguishedVariables()[k]);
			ObjectPropertyMember opm = KAON2Manager.factory().objectPropertyMember(op, refMaterializedIndividual ,i);
//			Literal llps =  KAON2Manager.factory().literal(true, op,refMaterializedIndividual,i);

			//Formula opm = KAON2Manager.factory().literal(true, op,refMaterializedIndividual,i);
		//	Rule opm = KAON2Manager.factory().rule(
		//			llps,                          // this is the rule head, i.e. the consequent of the rule
		 //           new Literal[] { }   // this is the rule body, i.e. the condition of the rule
		  //      );

					if (!kb.getOntology().containsAxiom(opm,false)) {
						changes.add(new OntologyChangeEvent(opm, OntologyChangeEvent.ChangeType.ADD));
						kb.getOntology().applyChanges(changes);
						changes.clear();

						rollback.add(new OntologyChangeEvent(opm, OntologyChangeEvent.ChangeType.REMOVE));
					}
			
		}
		q.next();
	}
	
	q.close();
	q.dispose();
	
	updateLastMaterializationDepth(connectedVars);

	return rollback;
}


/**
 * Materializes answers of the reference query
 * @param kb Knowledge base to which answers are to be materialized
 * @return List of changes (of type<OntologyChangeEvent>) to knowledge base, made by adding materialized answers, to rollback  
 * @throws InterruptedException 
 * @throws KAON2Exception 
 * 
*/
public List materializeResultsInit(KnowledgeBase kb) throws KAON2Exception, InterruptedException{
	//Pattern p = getRewrittenPattern().rewriteVariablesToDistinguishedOnes();

	System.out.println("matInit:"+"depth:"+getDepth());
	
	OWLClass refMaterializedClass =  KAON2Manager.factory().owlClass(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(getDepth()));
	//PredicateSymbol refMaterializedClass =  KAON2Manager.factory().predicateSymbol(kb.getOntology().getOntologyURI()+"#$ref_"+String.valueOf(getDepth()),1);

	List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
	List<OntologyChangeEvent> rollback=new ArrayList<OntologyChangeEvent>();

	Query q = kb.getReasoner().createQuery(getAtom().getAtom().getPredicate());
	q.open();

	while (!q.afterLast()) {
		Term t[] = q.tupleBuffer();
		
		Individual refMaterializedIndividual = (Individual) t[0];

		ClassMember cm = KAON2Manager.factory().classMember(refMaterializedClass,refMaterializedIndividual);
	    //Literal ll =  KAON2Manager.factory().literal(true, refMaterializedClass,refMaterializedIndividual);
		//Formula cm =  KAON2Manager.factory().literal(true, refMaterializedClass,refMaterializedIndividual);

//			Rule cm = KAON2Manager.factory().rule(
//				ll,                          // this is the rule head, i.e. the consequent of the rule
//	            new Literal[] { }   // this is the rule body, i.e. the condition of the rule
//	        );
		
	//	KAON2Manager.factory().groundTerm(arg0, arg1) (refMaterializedClass,refMaterializedIndividual);

		if (!kb.getOntology().containsAxiom(cm,false)) {
				changes.add(new OntologyChangeEvent(cm ,OntologyChangeEvent.ChangeType.ADD));
				kb.getOntology().applyChanges(changes);
				changes.clear();
				
				rollback.add(new OntologyChangeEvent(cm,OntologyChangeEvent.ChangeType.REMOVE));
		}
		
		q.next();
	}
	
	q.close();
	q.dispose();

	
	return rollback;
}

/**
 * Disposes resources associated to given node
 **/
public void dispose() {
	//getPattern().getQuery().dispose();
	setPattern(null);
		
	if (getRewrittenPattern()!=null){
			//System.out.println("usuwa rewritten");
		//getRewrittenPattern().getQuery().dispose();
		setRewrittenPattern(null);
	}
}

}
