/**
 * Created on 2008-03-19 09:34:15
 */
package org.put.semintec.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.text.SimpleDateFormat;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import jdsl.core.api.ObjectIterator;
import jdsl.core.api.Position;
import jdsl.core.api.PositionIterator;
import jdsl.core.api.Tree;

import org.put.semintec.core.KnowledgeBase;
import org.put.semintec.mining.association.Pattern;
import org.put.semintec.mining.association.AssociationRule;
import org.put.semintec.mining.datastructures.Node;
import org.put.semintec.mining.datastructures.Trie;
import org.put.semintec.utils.Statistics;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.Namespaces;
import org.semanticweb.kaon2.api.logic.Conjunction;
import org.semanticweb.kaon2.api.logic.Literal;
import org.semanticweb.kaon2.api.logic.Predicate;
import org.semanticweb.kaon2.api.logic.QueryDefinition;
import org.semanticweb.kaon2.api.logic.Variable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;


/**
 * Provides methods that perform I/O operations.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class IOUtils {

	/**
	 * Saves statistics about data mining execution into specified file
	 * @param params Parameters of data mining execution
	 * @param statistics Collected statistics
	 * @param fileName Name of the file to write statistics (with path) 
	*/
	public void saveStatistics(HashMap params, Statistics statistics, String fileName){
     	PrintWriter out;
		try {
			out = new PrintWriter(
			    new BufferedWriter(
			      new FileWriter(fileName, true)));
			
			if (statistics.getMiningTask() == Statistics.PATTERN){
 			  //info about the execution	
				//file name with execution setup
				out.print(params.get("setupfile")+";");
				//main ontology URI
				out.print(params.get("ontologyURI")+";");
				//reference class
				out.print(params.get("refclass")+";");
				//minimum support
				out.print(params.get("minsup")+";");
				//max length
		        out.print(params.get("maxlength")+";");
		        //answer caching on?
		        out.print(params.get("answercaching")+";");
		       
		      //statistics    
		        //candidates and frequent patterns
		        out.print(statistics.getNumCandidatesUpToMaxLength()+";");
		        out.print(statistics.getNumPatternsUpToMaxLength()+";");
		        out.print(statistics.getNumCandidatesWithMaxLength()+";");
		        out.print(statistics.getNumPatternsWithMaxLength()+";");

		        //generated, satisfiable and sfree patterns
		        out.print(statistics.getNumGENUpToMaxLength()+";");
		        out.print(statistics.getNumGENWithMaxLength()+";");
		        out.print(statistics.getNumSATUpToMaxLength()+";");
		        out.print(statistics.getNumSATWithMaxLength()+";");
		        out.print(statistics.getNumSFREEUpToMaxLength()+";");
		        out.print(statistics.getNumSFREEWithMaxLength()+";");

		        //execution times
		        SimpleDateFormat sdf = new SimpleDateFormat("MM/dd/yyyy HH:mm:ss");

		        out.print(sdf.format(statistics.getStartDate())+";");
		        out.print(sdf.format(statistics.getStopDate())+";");
		        out.print(statistics.calculateExecutionTime()+";");
			}
	     	out.println();
	     	out.close();		

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Saves patterns, discovered during data mining execution, into specified file
	 * @param params Parameters of data mining execution
	 * @param trie Trie structure that stores frequent patterns
	 * @param fileName Name of the file to write statistics (with path) 
	*/
	public void savePatterns(HashMap params, Trie trie, String fileName){
     	PrintWriter out;
		try {
			out = new PrintWriter(
			    new BufferedWriter(
			      new FileWriter(fileName, true)));
		  //info about the execution	
			//file name with execution setup
			out.print(params.get("setupfile")+";");
			//main ontology URI
			out.print(params.get("ontologyURI")+";");
			//reference class
			out.print(params.get("refclass")+";");
			//minimum support
			out.print(params.get("minsup")+";");
			//max length
	        out.print(params.get("maxlength")+";");
	        //answer caching on?
	        out.print(params.get("answercaching")+";");
		    out.println();
	       
		  //patterns
			dumpPatternsFromTrie(trie, trie.getTree().root(), out);

		    out.println();
	     	out.close();		

		} catch (IOException e) {
			// TODO Auto-generated catch blocke
			e.printStackTrace();
		} catch (KAON2Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Saves rules, generated from discovered patterns, into specified file
	 * @param params Parameters of data mining execution
	 * @param rules List that stores frequent patterns
	 * @param fileName Name of the file to write statistics (with path) 
	*/
	public void saveRules(HashMap params, ArrayList rules, String fileName){
     	PrintWriter out;
		try {
			out = new PrintWriter(
			    new BufferedWriter(
			      new FileWriter(fileName, true)));
		  //info about the execution	
			//file name with execution setup
			out.print(params.get("setupfile")+";");
			//main ontology URI
			out.print(params.get("ontologyURI")+";");
			//reference class
			out.print(params.get("refclass")+";");
			//minimum support
			out.print(params.get("minsup")+";");
			//max length
	        out.print(params.get("maxlength")+";");
			//minimum confidence
			out.print(params.get("minconf")+";");
	        //answer caching on?
	        out.print(params.get("answercaching")+";");
		    out.println();
	       
		  //rules
			dumpRules(rules, out);

		    out.println();
	     	out.close();		

		} catch (IOException e) {
			// TODO Auto-generated catch blocke
			e.printStackTrace();
		} catch (KAON2Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Writes patterns stored in a trie to specified output
	 * 
	 * @param trie Trie to dump to specified output
	 * @param parent Position of currently handled node in the trie 
	 * @param out Output where patterns should be dumped 
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/

	private void dumpPatternsFromTrie(Trie trie, Position parent, PrintWriter out) throws KAON2Exception, InterruptedException{
		if (trie.getTree().isInternal(parent)) {
			Position child = trie.getTree().firstChild(parent);

			//dump node
			out.print(((Node)child.element()).getPattern().getQueryDefinition().getFormula().toString()+";");
			out.print(((Node)child.element()).getDepth()+";");
			out.print(((Node)child.element()).getPattern().getFrequency()+";");
			out.print(((Node)child.element()).getPattern().getSupport());
			out.println();
			
			dumpPatternsFromTrie(trie, child, out);

			while (child != trie.getTree().lastChild(parent)) {
				child = trie.getTree().siblingAfter(child);
				//dump node
				out.print(((Node)child.element()).getPattern().getQueryDefinition().getFormula().toString()+";");
				out.print(((Node)child.element()).getDepth()+";");
				out.print(((Node)child.element()).getPattern().getFrequency()+";");
				out.print(((Node)child.element()).getPattern().getSupport());
				out.println();

				dumpPatternsFromTrie(trie, child, out); 
			}
		}
	}

	/**
	 * Dumps rules from list to specified output
	 * 
	 * @param rules List of rules to dump to specified output
	 * @param out Output where patterns should be dumped 
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/

	private void dumpRules(ArrayList rules, PrintWriter out) throws KAON2Exception, InterruptedException{
		for (Iterator i = rules.iterator(); i.hasNext();) {
			AssociationRule ar = (AssociationRule) i.next();

			//dump node
			out.print(ar.getAntecedent().getQueryDefinition().getFormula().toString()+"=>");
			out.print(ar.getConsequent().getQueryDefinition().getFormula().toString()+";");
			//out.print(ar.getAntecedent().getFrequency()+";");
			//out.print(ar.getAntecedent().getSupport());
			//out.print(ar.getConfidence());

			out.println();
		}
	}

	
	/**
	 * Writes a given trie data structure to specified file in GraphML format.
	 * @param trie A given trie.
	 * @param Name of a file to write (with path)
	 * @param params Specify what information from a trie should be written to a file.  
	 * @throws IOException 
	*/
	public void saveToGraphML(Trie trie, String fileName, String[] params) throws IOException{
	{
		int edgeCounter = 0;
		
     	PrintWriter out =
	      new PrintWriter(
	        new BufferedWriter(
	          new FileWriter(fileName)));
     	
     	out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
     	out.println("		<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns/graphml\"  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
     	out.println("		xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns/graphml http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd\" xmlns:y=\"http://www.yworks.com/xml/graphml\">");
     	out.println("		  <key id=\"d0\" for=\"node\" yfiles.type=\"nodegraphics\"/>");
     	out.println("		  <key id=\"d1\" for=\"edge\" yfiles.type=\"edgegraphics\"/>");
     	out.println("		  <graph id=\"G\" edgedefault=\"directed\">");
     	
		for (PositionIterator i=trie.getTree().positions(); i.hasNext();)
		{
			Position p = (Position) i.nextPosition();
			Node n = (Node) p.element();
			
			out.println("<node id=\"n"+n.toString()+"\">");
			out.println("  <data key=\"d0\" >");
			out.println("    <y:ShapeNode>");
			out.println("      <y:Geometry  x=\"183.0\" y=\"1.0\" width=\"70.0\" height=\"30.0\"/>");
			out.println("      <y:Fill color=\"#FFFFFF\"  transparent=\"false\"/>");
			out.println("      <y:BorderStyle type=\"line\" width=\"1.0\" color=\"#000000\" />");
			String info = n.getAtom().getAtom().toString();
			//TODO- tmp
			info=info.replaceAll("]", "");
			if (info.contains("#")){
				info=info.substring(info.indexOf("#")+1);
			}
			info+=" "+n.how;
			out.println("      <y:NodeLabel x=\"4.5\" y=\"5.6494140625\" width=\"61.0\" height=\"18.701171875\" visible=\"true\" alignment=\"center\" fontFamily=\"Dialog\" fontSize=\"12\" fontStyle=\"plain\" textColor=\"#000000\" modelName=\"internal\" modelPosition=\"c\" autoSizePolicy=\"content\">"+info+"</y:NodeLabel>");
			out.println("      <y:Shape type=\"rectangle\"/>");
			out.println("    </y:ShapeNode>");
			out.println("  </data>");
			out.println("</node>");
			
			//iterate over children of the node if any and add edges
			if (trie.getTree().isInternal(p)){
				Position pChild = trie.getTree().firstChild(p);
				//add edge between parent and child
				Node nChild = (Node)pChild.element(); 
				
				out.println("<edge id=\"e"+edgeCounter+"\" source=\"n"+n.toString()+"\" target=\"n"+nChild.toString()+"\">");
				out.println("  <data key=\"d1\" >");
				out.println("    <y:PolyLineEdge>");
				out.println("      <y:Path sx=\"0.0\" sy=\"0.0\" tx=\"0.0\" ty=\"0.0\"/>");
				out.println("      <y:LineStyle type=\"line\" width=\"1.0\" color=\"#000000\" />");
				out.println("      <y:Arrows source=\"none\" target=\"standard\"/>");
				out.println("      <y:BendStyle smoothed=\"false\"/>");
				out.println("    </y:PolyLineEdge>");
				out.println("  </data>");
				out.println("</edge>");
				
				while (pChild != trie.getTree().lastChild(p)) {
					pChild = trie.getTree().siblingAfter(pChild);
					//add edge between parent and child	
					nChild = (Node)pChild.element(); 
					
					out.println("<edge id=\"e"+edgeCounter+"\" source=\"n"+n.toString()+"\" target=\"n"+nChild.toString()+"\">");
					out.println("  <data key=\"d1\" >");
					out.println("    <y:PolyLineEdge>");
					out.println("      <y:Path sx=\"0.0\" sy=\"0.0\" tx=\"0.0\" ty=\"0.0\"/>");
					out.println("      <y:LineStyle type=\"line\" width=\"1.0\" color=\"#000000\" />");
					out.println("      <y:Arrows source=\"none\" target=\"standard\"/>");
					out.println("      <y:BendStyle smoothed=\"false\"/>");
					out.println("    </y:PolyLineEdge>");
					out.println("  </data>");
					out.println("</edge>");
				}

				edgeCounter++;
			}
		}

			

		out.println("  </graph>");
		out.println(" </graphml>");
		out.close();		
		}
	}
	
	
	/**
	 * Writes a given taxonomy of predicates or nodes  to specified file in GraphML format.
	 * @param taxonomy A given taxonomy
	 * @param Name of a file to write (with path)
	 * @param params Specify what information from a taxonomy should be written to a file.  
	 * @throws IOException 
	*/
	public void saveToGraphML(Tree taxonomy, String fileName, String[] params) throws IOException{
	{
		int edgeCounter = 0;
		
     	PrintWriter out =
	      new PrintWriter(
	        new BufferedWriter(
	          new FileWriter(fileName)));
     	
     	out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
     	out.println("		<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns/graphml\"  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
     	out.println("		xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns/graphml http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd\" xmlns:y=\"http://www.yworks.com/xml/graphml\">");
     	out.println("		  <key id=\"d0\" for=\"node\" yfiles.type=\"nodegraphics\"/>");
     	out.println("		  <key id=\"d1\" for=\"edge\" yfiles.type=\"edgegraphics\"/>");
     	out.println("		  <graph id=\"G\" edgedefault=\"directed\">");
     	
		for (PositionIterator i=taxonomy.positions(); i.hasNext();)
		{
			Position p = (Position) i.nextPosition();
			Object o = p.element();
			Object value = o;
			if (o instanceof Predicate){
				value = (Predicate) o; 
			} else if (o instanceof Node) {
				Node n = (Node) o;
				value = n.getAtom().getAtom();
			} 
			
			out.println("<node id=\"n"+value+"\">");
			out.println("  <data key=\"d0\" >");
			out.println("    <y:ShapeNode>");
			out.println("      <y:Geometry  x=\"183.0\" y=\"1.0\" width=\"70.0\" height=\"30.0\"/>");
			out.println("      <y:Fill color=\"#FFFFFF\"  transparent=\"false\"/>");
			out.println("      <y:BorderStyle type=\"line\" width=\"1.0\" color=\"#000000\" />");
//			String info = (String)value;

//			if (info.contains("#")){
//				info=info.substring(info.indexOf("#")+1);
//			}
			//info+=" "+n.how;
			out.println("      <y:NodeLabel x=\"4.5\" y=\"5.6494140625\" width=\"61.0\" height=\"18.701171875\" visible=\"true\" alignment=\"center\" fontFamily=\"Dialog\" fontSize=\"12\" fontStyle=\"plain\" textColor=\"#000000\" modelName=\"internal\" modelPosition=\"c\" autoSizePolicy=\"content\">"+value+"</y:NodeLabel>");
			out.println("      <y:Shape type=\"rectangle\"/>");
			out.println("    </y:ShapeNode>");
			out.println("  </data>");
			out.println("</node>");
			
			//iterate over children of the node if any and add edges
			if (taxonomy.isInternal(p)){
				Position pChild = taxonomy.firstChild(p);
				//add edge between parent and child
//				Predicate predChild = (Predicate)pChild.element(); 
				Object oChild = pChild.element();
				Object valueChild = oChild;
				if (oChild instanceof Predicate){
					valueChild = (Predicate) oChild; 
				} else if (oChild instanceof Node) {
					Node n = (Node) oChild;
					valueChild = n.getAtom().getAtom();
				}

				
				out.println("<edge id=\"e"+edgeCounter+"\" source=\"n"+value+"\" target=\"n"+valueChild+"\">");
				out.println("  <data key=\"d1\" >");
				out.println("    <y:PolyLineEdge>");
				out.println("      <y:Path sx=\"0.0\" sy=\"0.0\" tx=\"0.0\" ty=\"0.0\"/>");
				out.println("      <y:LineStyle type=\"line\" width=\"1.0\" color=\"#000000\" />");
				out.println("      <y:Arrows source=\"none\" target=\"standard\"/>");
				out.println("      <y:BendStyle smoothed=\"false\"/>");
				out.println("    </y:PolyLineEdge>");
				out.println("  </data>");
				out.println("</edge>");
				
				while (pChild != taxonomy.lastChild(p)) {
					pChild = taxonomy.siblingAfter(pChild);
					//add edge between parent and child	
					oChild = pChild.element();
					valueChild = oChild;
					if (oChild instanceof Predicate){
						valueChild = (Predicate) oChild; 
					} else if (oChild instanceof Node) {
						Node n = (Node) oChild;
						valueChild = n.getAtom().getAtom();
					}
					
					out.println("<edge id=\"e"+edgeCounter+"\" source=\"n"+value+"\" target=\"n"+valueChild+"\">");
					out.println("  <data key=\"d1\" >");
					out.println("    <y:PolyLineEdge>");
					out.println("      <y:Path sx=\"0.0\" sy=\"0.0\" tx=\"0.0\" ty=\"0.0\"/>");
					out.println("      <y:LineStyle type=\"line\" width=\"1.0\" color=\"#000000\" />");
					out.println("      <y:Arrows source=\"none\" target=\"standard\"/>");
					out.println("      <y:BendStyle smoothed=\"false\"/>");
					out.println("    </y:PolyLineEdge>");
					out.println("  </data>");
					out.println("</edge>");
				}

				edgeCounter++;
			}
		}

			

		out.println("  </graph>");
		out.println(" </graphml>");
		out.close();		
		}
	}
	
	
	/**
	 * Reads input arguments for the mining application, stored in XML file 
	 * @param fileName Name of the file to read (with path)
     * @return HasSet with application input arguments
	*/

	public HashMap readXMLSetupFile(String fileName) {
	    HashMap params = new HashMap();
		  params.put("setupfile", fileName);

		try {
		  File file = new File(fileName);
		  DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		  DocumentBuilder db = dbf.newDocumentBuilder();
		  Document doc = db.parse(file);
		  doc.getDocumentElement().normalize();
		  
		  //System.out.println("Root element " + doc.getDocumentElement().getNodeName());
          
		  NodeList nLst = doc.getElementsByTagName("ontology");
          Element ontology = (Element) nLst.item(0);
          
          nLst = ontology.getElementsByTagName("physicalURI");
          Element physicalURI = (Element) nLst.item(0);
          params.put("physicalURI", physicalURI.getTextContent());
          nLst = ontology.getElementsByTagName("ontologyURI");
          Element ontologyURI = (Element) nLst.item(0);
          params.put("ontologyURI", ontologyURI.getTextContent());
          
          //TODO - imported ontologies
		
          nLst = doc.getElementsByTagName("refclass");
          Element refclass = (Element) nLst.item(0);
          params.put("refclass", refclass.getTextContent());

          nLst = doc.getElementsByTagName("minsup");
          Element minsup = (Element) nLst.item(0);
          params.put("minsup", minsup.getTextContent());
          
          nLst = doc.getElementsByTagName("maxlength");
          Element maxlength = (Element) nLst.item(0);
          params.put("maxlength", maxlength.getTextContent());
          
          nLst = doc.getElementsByTagName("outputFileNames");
          Element outputFileNames = (Element) nLst.item(0);
          nLst = outputFileNames.getElementsByTagName("patternsFileName");
          Element patternsFileName = (Element) nLst.item(0);
          params.put("patternsFileName", patternsFileName.getTextContent());
          nLst = outputFileNames.getElementsByTagName("rulesFileName");
          Element rulesFileName = (Element) nLst.item(0);
          if (rulesFileName!=null){
        	  params.put("rulesFileName", rulesFileName.getTextContent());
          }
          nLst = outputFileNames.getElementsByTagName("statisticsFileName");
          Element statisticsFileName = (Element) nLst.item(0);
          params.put("statisticsFileName", statisticsFileName.getTextContent());
 
          nLst = doc.getElementsByTagName("answercaching");
          Element answercaching = (Element) nLst.item(0);
          params.put("answercaching", answercaching.getTextContent());
          
		  nLst = doc.getElementsByTagName("bias");
          Element bias = (Element) nLst.item(0);
          nLst = bias.getElementsByTagName("predicates");
          Element predicates = (Element) nLst.item(0);
          params.put("predicates", predicates.getTextContent());

          ArrayList preds = new ArrayList();
          nLst = predicates.getElementsByTagName("predicate");
          for (int i=0; i<nLst.getLength();i++){
        	  Element predicate = (Element) nLst.item(i);
        	  preds.add(predicate.getTextContent());
          }
          params.put("predlist", preds);
          
          //TODO - no such element exception
                   
	  }
	  catch (Exception e) {
	    e.printStackTrace();
	  }
      return params;
	} 
	
	/**
	 * Reads a list of patterns (conjunctive queries) from file  
	 * @param fileName Name of the file to read (with path)
	 * @param keyVar Key variable (the one in reference concept)
     * @return List of patterns
	*/
	public ArrayList readPatternList(String fileName, Variable keyVar){
		ArrayList patterns = new ArrayList();
		String line;
		BufferedReader br;

		Namespaces namespaces=new Namespaces();
//        namespaces.registerPrefix("example04","http://kaon2.semanticweb.org/example04#");
		
		try {
			br = new BufferedReader(new FileReader(fileName));
			while ((line = br.readLine()) != null){
				line=line.replaceAll("-1", "v1");
				line=line.replaceAll("2000", "v2");
				line=line.replaceAll("3000", "v3");
				line=line.replaceAll("4000", "v4");
				line=line.replaceAll("5000", "v5");

				//System.out.println(line);
				//make a pattern from the String
				Conjunction c = (Conjunction)KAON2Manager.factory().axiom(line, namespaces);
				QueryDefinition qd = (QueryDefinition)KAON2Manager.factory().queryDefinition(c, keyVar);
				//System.out.println("hihi:"+qd);
				Pattern p = new Pattern(qd);
				//add to patterns list
				patterns.add(p);	
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch(IOException e){
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (KAON2Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
		}

		return patterns;
	}
	
}