/**
 * Created on 2008-07-18 11:16:15
 */
package org.put.semintec.utils;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.put.semintec.core.KnowledgeBase;
import org.semanticweb.kaon2.api.Axiom;
import org.semanticweb.kaon2.api.Cursor;
import org.semanticweb.kaon2.api.Entity;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.OntologyChangeEvent;
import org.semanticweb.kaon2.api.Request;
import org.semanticweb.kaon2.api.owl.axioms.ClassMember;
import org.semanticweb.kaon2.api.owl.axioms.DataPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.Declaration;
import org.semanticweb.kaon2.api.owl.axioms.DifferentIndividuals;
import org.semanticweb.kaon2.api.owl.axioms.NegativeDataPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.NegativeObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.ObjectPropertyMember;
import org.semanticweb.kaon2.api.owl.axioms.SameIndividual;
import org.semanticweb.kaon2.api.owl.elements.Description;
import org.semanticweb.kaon2.api.owl.elements.Individual;
import org.semanticweb.kaon2.api.owl.elements.ObjectProperty;

/**
 * Provides methods that manipulate the original KB.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class KBUtils {
	/**
	 * Multiplies ABox of a given knowledge base <tt>k</tt> times
	 * @param kb Knowledge base
	 * @param k How many times ABox should be multiplied
	 * @return Knowledge base with multiplied ABox
	 * //TODO unfinished - only class member and object property member assertions are multiplied 
	*/

	public KnowledgeBase multiplyABox(KnowledgeBase kb, int k) {
		KnowledgeBase multipliedKB = null;
		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();

		try {
			multipliedKB = kb.copy(true);
		
			//iterate over ABox assertions
			Request<Axiom> axiomRequest=multipliedKB.getOntology().createAxiomRequest();
			Cursor<Axiom> cursor=axiomRequest.openCursor();
			ArrayList axioms = new ArrayList();
			while (cursor.hasNext()) {
				Axiom axiom=cursor.next();
				axioms.add(axiom);
			}
			cursor.close();
			for (Iterator it = axioms.iterator(); it.hasNext();){
				Axiom axiom=(Axiom)it.next();

			  if (axiom instanceof ClassMember || 
					axiom instanceof ObjectPropertyMember ||
					axiom instanceof NegativeObjectPropertyMember ||
					axiom instanceof DataPropertyMember ||
					axiom instanceof NegativeDataPropertyMember ||
					axiom instanceof DifferentIndividuals ||
					axiom instanceof SameIndividual ||
					(axiom instanceof Declaration && ((Declaration) axiom).getEntity() instanceof Individual)){ 
						for (int i=0; i<k-1; i++) {
							//for each class assertion create k-1 copies of it
							if (axiom instanceof ClassMember) {
								Description desc = (Description)((ClassMember)axiom).getDescription();
								Individual iCopy = KAON2Manager.factory().individual(((ClassMember)axiom).getIndividual().getURI()+"$"+i);
								ClassMember cm = KAON2Manager.factory().classMember(desc, iCopy);
								if (!(multipliedKB.getOntology().containsAxiom(cm, true))){
									changes.add(new OntologyChangeEvent(cm,OntologyChangeEvent.ChangeType.ADD));
									multipliedKB.getOntology().applyChanges(changes);
									changes.clear();
								}
							} else if (axiom instanceof ObjectPropertyMember) {
								//for each property assertion create k-1 copies of it
								ObjectProperty op = (ObjectProperty) ((ObjectPropertyMember)axiom).getObjectProperty();
								Individual iCopySource = KAON2Manager.factory().individual(((ObjectPropertyMember)axiom).getSourceIndividual().getURI()+"$"+i);
								Individual iCopyTarget = KAON2Manager.factory().individual(((ObjectPropertyMember)axiom).getTargetIndividual().getURI()+"$"+i);
								ObjectPropertyMember opm = KAON2Manager.factory().objectPropertyMember(op, iCopySource, iCopyTarget);
								if (!(multipliedKB.getOntology().containsAxiom(opm, true))){
									changes.add(new OntologyChangeEvent(opm,OntologyChangeEvent.ChangeType.ADD));
									multipliedKB.getOntology().applyChanges(changes);
									changes.clear();
								}
							}
							//TODO the other ABox axioms
						}
				}
			}
		} catch (KAON2Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

        return multipliedKB;
	}
}
