/**
 * Created on 2008-08-22 16:20:15
 */

package org.put.semintec.utils;

import java.util.ArrayList;
import java.util.Iterator;

import org.put.semintec.core.KnowledgeBase;
import org.put.semintec.mining.association.Pattern;
import org.semanticweb.kaon2.api.KAON2Exception;

/**
 * Provides methods that manipulate a set of patterns.
 * @author Agnieszka Lawrynowicz
 * 
 */

public class MiningUtils {
	/**
	 * Selects only s-free patterns, w.r.t kb, from the given list of patterns
	 * @param patterns List of patterns to filter
	 * @kb Knowledge base
	 * @return the list of s-free patterns 
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	 * 
	 */
	public ArrayList selectSFree(ArrayList patterns, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		ArrayList sFreePatterns = new ArrayList();
		for (Iterator i = patterns.iterator(); i.hasNext();){
			Pattern p  = (Pattern) i.next();
			if (p.isSFree(kb)){
				sFreePatterns.add(p);
			}
		}
		return sFreePatterns;
	}
	
	/**
	 * Selects only one pattern from each equivalence class, w.r.t kb, from the given list of patterns
	 * @param patterns List of patterns to filter
	 * @kb Knowledge base
	 * @return the list of patterns selected as equivalence class representative
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	 * 
	 */
	public ArrayList selectOnePatternPerEquivalenceClass(ArrayList patterns, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		ArrayList uniquePatterns = new ArrayList();
		for (Iterator i = patterns.iterator(); i.hasNext();){
			Pattern p  = (Pattern) i.next();
			boolean notFound = true;
			for (Iterator ii = uniquePatterns.iterator(); ii.hasNext();){
				Pattern p2  = (Pattern) ii.next();
				if (p.equivalentTo(p2, kb)){
					if (!(p.getQueryDefinition().toString().compareTo(p2.getQueryDefinition().toString())==0)){
						System.out.println("p1:"+p.getQueryDefinition()+"p2:"+p2.getQueryDefinition());
					}
					notFound = false;
				}
			}
			if (notFound){
				uniquePatterns.add(p);
			}
		}
		return uniquePatterns;
	}
}
